@@ -633,11 +633,16 @@ SV *p5_call_package_method(PerlInterpreter *my_perl, char *package, char *name,
     }
 }
 
-GV *p5_look_up_package_method(PerlInterpreter *my_perl, char *module, char *name) {
+GV *p5_look_up_package_method(PerlInterpreter *my_perl, char *module, char *name, I32 local) {
     PERL_SET_CONTEXT(my_perl);
     {
         HV * const pkg = gv_stashpvn(module, strlen(module), 0);
-        GV * const gv = gv_fetchmeth_pvn_autoload(pkg, name, strlen(name), -1, SVf_UTF8);
+        GV * gv = gv_fetchmeth_pvn_autoload(pkg, name, strlen(name), -1, SVf_UTF8);
+        if (gv && local) {
+            GV * const super_gv = gv_fetchmeth_pvn_autoload(pkg, name, strlen(name), -1, SVf_UTF8 | GV_SUPER);
+            if (super_gv && (super_gv == gv || GvCV(super_gv) == GvCV(gv)))
+                gv = NULL;
+        }
         if (gv && isGV(gv))
             return gv;
         return NULL;
@@ -667,7 +672,7 @@ SV *p5_call_inherited_package_method(PerlInterpreter *my_perl, char *package, ch
         XPUSHs(sv_2mortal(package_sv));
         push_arguments(sp, len, args);
 
-        GV * const gv = p5_look_up_package_method(my_perl, base_package, name);
+        GV * const gv = p5_look_up_package_method(my_perl, base_package, name, 0);
         SV * const rv = sv_2mortal(newRV_inc((SV*)GvCV(gv)));
         *count = call_sv(rv, flags);
 
@@ -684,17 +689,6 @@ SV *p5_call_inherited_package_method(PerlInterpreter *my_perl, char *package, ch
     }
 }
 
-GV *p5_look_up_method(PerlInterpreter *my_perl, SV *obj, char *name) {
-    PERL_SET_CONTEXT(my_perl);
-    {
-        HV * const pkg = SvSTASH((SV*)SvRV(obj));
-        GV * const gv = Perl_gv_fetchmethod_autoload(aTHX_ pkg, name, TRUE);
-        if (gv && isGV(gv))
-            return gv;
-        return NULL;
-    }
-}
-
 char *p5_stash_name(PerlInterpreter *my_perl, SV *obj) {
     HV * const pkg = SvSTASH((SV*)SvRV(obj));
     return HvNAME(pkg);
@@ -1202,25 +1196,6 @@ MGVTBL p5_inline_hash_mg_vtbl = {
     0x0
 };
 
-void p5_rebless_object(PerlInterpreter *my_perl, SV *obj, char *package, IV i) {
-    PERL_SET_CONTEXT(my_perl);
-    {
-        SV * const inst = SvRV(obj);
-        HV *stash = gv_stashpv(package, GV_ADD);
-        if (stash == NULL)
-            croak("Perl6::Object not found!? Forgot to call init_callbacks?");
-        (void)sv_bless(obj, stash);
-
-        _perl6_magic priv;
-
-        /* set up magic */
-        priv.key = PERL6_MAGIC_KEY;
-        priv.index = i;
-        priv.is_wrapper = 0;
-        sv_magicext(inst, inst, PERL_MAGIC_ext, &p5_inline_mg_vtbl, (char *) &priv, sizeof(priv));
-    }
-}
-
 SV *p5_add_magic(PerlInterpreter *my_perl, SV *inst, IV i) {
     PERL_SET_CONTEXT(my_perl);
     {
@@ -1431,22 +1406,34 @@ AV *create_args_array(const I32 ax, I32 items, I32 num_fixed_args) {
 }
 
 void return_retval(const I32 ax, SV **sp, SV *retval) {
-    if (GIMME_V == G_VOID) {
+    if (retval == NULL || GIMME_V == G_VOID) {
         XSRETURN_EMPTY;
     }
     if (GIMME_V == G_ARRAY) {
-        AV* const av = (AV*)SvRV(retval);
-        I32 const len = av_len(av) + 1;
-        I32 i;
-        for (i = 0; i < len; i++) {
-            XPUSHs(sv_2mortal(av_shift(av)));
+        if (SvROK(retval) && SvTYPE(SvRV(retval)) == SVt_PVAV) {
+            AV* const av = (AV*)SvRV(retval);
+            I32 const len = av_len(av) + 1;
+            I32 i;
+            for (i = 0; i < len; i++) {
+                XPUSHs(sv_2mortal(av_shift(av)));
+            }
+            XSRETURN(len);
+        }
+        else {
+            XPUSHs(retval);
+            XSRETURN(1);
         }
-        XSRETURN(len);
     }
     else {
-        AV* const av = (AV*)SvRV(retval);
-        XPUSHs(sv_2mortal(av_shift(av)));
-        XSRETURN(1);
+        if (SvROK(retval) && SvTYPE(SvRV(retval)) == SVt_PVAV) {
+            AV* const av = (AV*)SvRV(retval);
+            XPUSHs(sv_2mortal(av_shift(av)));
+            XSRETURN(1);
+        }
+        else {
+            XPUSHs(retval);
+            XSRETURN(1);
+        }
     }
 }
 
@@ -1513,7 +1500,14 @@ XS(p5_destroy_p5_object) {
     if (mg) {
         _perl6_magic* const p6mg = (_perl6_magic*) mg->mg_ptr;
         /* need to be extra careful here as PL_modglobal could have been cleaned already */
-        if (p6mg->index != -1 && !PL_in_clean_objs) {
+        if (
+            (      p6mg->key == PERL6_MAGIC_KEY
+                || p6mg->key == PERL6_HASH_MAGIC_KEY
+                || p6mg->key == PERL6_EXTENSION_MAGIC_KEY
+            )
+            && p6mg->index != -1
+            && !PL_in_clean_objs
+        ) {
             SV **cbs_entry = hv_fetchs(PL_modglobal, "Inline::Perl5 callbacks", 0);
             if (cbs_entry) {
                 perl6_callbacks *cbs = (perl6_callbacks*)SvIV(*cbs_entry);
